 "======================================================================
|
|   Class ScaledDecimal Definitions
|
|   $Revision: 1.95.1$
|   $Date: 2000/12/27 10:45:49$
|   $Author: pb$
|
 ======================================================================"


"======================================================================
|
| Copyright 1988-92, 1994-95, 1999, 2000 Free Software Foundation, Inc.
| Written by Paolo Bonzini.
|
| This file is part of the GNU Smalltalk class library.
|
| The GNU Smalltalk class library is free software; you can redistribute it
| and/or modify it under the terms of the GNU Lesser General Public License
| as published by the Free Software Foundation; either version 2.1, or (at
| your option) any later version.
| 
| The GNU Smalltalk class library is distributed in the hope that it will be
| useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
| General Public License for more details.
| 
| You should have received a copy of the GNU Lesser General Public License
| along with the GNU Smalltalk class library; see the file COPYING.LESSER.
| If not, write to the Free Software Foundation, 59 Temple Place - Suite
| 330, Boston, MA 02111-1307, USA.  
|
 ======================================================================"


Number subclass:  #ScaledDecimal
	instanceVariableNames: 'fraction scale '
	classVariableNames: 'Zero One'
	poolDictionaries: ''
	category: 'Kernel-Numbers'!

ScaledDecimal comment: 
'ScaledDecimal provides a numeric representation of fixed point decimal
numbers able to accurately represent decimal fractions.  It supports
unbounded precision, with no limit to the number of digits before and
after the decimal point.'!

!ScaledDecimal class methodsFor: 'instance creation'!

newFromNumber: aNumber scale: scale
    "Answer a new instance of ScaledDecimal, representing a decimal
     fraction with a decimal representation considered valid up to the
     scale-th digit."
    ^(self basicNew)
	setFraction: aNumber asFraction scale: scale;
	yourself! !

!ScaledDecimal class methodsFor: 'constants'!

initialize
    "Initialize the receiver's class variables"

    "We cannot create ScaledDecimal with 1.0s0 during the very
     first phases of bootstrapping, because the lexer calls in to
     #asScaledDecimal: -- but it is not a problem."
    One := self newFromNumber: 1 scale: 0.
    Zero := self newFromNumber: 0 scale: 0! !

!ScaledDecimal methodsFor: 'constants'!

one
    "Answer the receiver's representation of one."
    ^One!

zero
    "Answer the receiver's representation of zero."
    ^Zero! !

!ScaledDecimal methodsFor: 'coercion'!

fractionPart
    "Answer the fractional part of the receiver."
    ^ScaledDecimal newFromNumber: fraction fractionPart scale: scale!

integerPart
    "Answer the fractional part of the receiver."
    ^ScaledDecimal newFromNumber: fraction integerPart scale: scale!

truncated
    "Answer the receiver, converted to an Integer and truncated towards
     -infinity."
    ^fraction truncated!

asFloat
    "Answer the receiver, converted to a Float"
    ^fraction asFloat!

asFraction
    "Answer the receiver, converted to a Fraction"
    ^fraction!

coerce: aNumber
    "Answer aNumber, converted to a ScaledDecimal with the same scale
     as the receiver."
    ^ScaledDecimal
	newFromNumber: aNumber asFraction
	scale: scale
!

generality
    "Return the receiver's generality"
    ^250
! !

!ScaledDecimal methodsFor: 'arithmetic'!

+ aNumber 
    "Sum two numbers and answer the result."

    (aNumber generality = self generality)
	ifTrue: [^ScaledDecimal
	    newFromNumber: fraction + aNumber asFraction
	    scale: (scale max: aNumber scale) ]
	ifFalse: [^self retrySumCoercing: aNumber]!

- aNumber 
     "Subtract aNumber from the receiver and answer the result."

    (aNumber generality = self generality)
	ifTrue: [^ScaledDecimal
	    newFromNumber: fraction - aNumber asFraction
	    scale: (scale max: aNumber scale) ]
	ifFalse: [^self retryDifferenceCoercing: aNumber]!

* aNumber 
     "Multiply two numbers and answer the result."

    (aNumber generality = self generality)
	ifTrue: [^ScaledDecimal
	    newFromNumber: fraction * aNumber asFraction
	    scale: (scale max: aNumber scale) ]
	ifFalse: [^self retryMultiplicationCoercing: aNumber]!


/ aNumber 
     "Divide two numbers and answer the result."

    (aNumber generality = self generality)
	ifTrue: [^ScaledDecimal
	    newFromNumber: fraction / aNumber asFraction
	    scale: (scale max: aNumber scale) ]
	ifFalse: [^self retryDivisionCoercing: aNumber]!

\\ aNumber 
    "Answer the remainder after integer division the receiver by aNumber 
    with truncation towards negative infinity."

    (aNumber generality = self generality)
	ifTrue: [^ScaledDecimal
	    newFromNumber: fraction \\ aNumber asFraction
	    scale: (scale max: aNumber scale) ]
	ifFalse: [^self retry: #\\ coercing: aNumber]!

// aNumber 
    "Answer the integer quotient after dividing the receiver by aNumber 
    with truncation towards negative infinity."
    ^fraction // aNumber! !

!ScaledDecimal methodsFor: 'comparing'!

< aNumber 
     "Answer whether the receiver is less than arg."

    (aNumber generality = self generality)
	ifTrue: [^fraction < aNumber asFraction ]
	ifFalse: [^self retryRelationalOp: #< coercing: aNumber]!

<= aNumber 
     "Answer whether the receiver is less than or equal to arg."

    (aNumber generality = self generality)
	ifTrue: [^fraction <= aNumber asFraction ]
	ifFalse: [^self retryRelationalOp: #<= coercing: aNumber]!

> aNumber 
     "Answer whether the receiver is greater than arg."

    (aNumber generality = self generality)
	ifTrue: [^fraction > aNumber asFraction ]
	ifFalse: [^self retryRelationalOp: #> coercing: aNumber]!

>= aNumber 
     "Answer whether the receiver is greater than or equal to arg."

    (aNumber generality = self generality)
	ifTrue: [^fraction >= aNumber asFraction ]
	ifFalse: [^self retryRelationalOp: #>= coercing: aNumber]!

= arg 
     "Answer whether the receiver is equal to arg."

    (arg isKindOf: Number) ifFalse: [^false].

    (arg generality = self generality)
	ifTrue: [^fraction = arg asFraction ]
	ifFalse: [^self retryEqualityCoercing: arg]!

~= arg 
     "Answer whether the receiver is not equal arg."

    (arg isKindOf: Number) ifFalse: [^true].

    (arg generality = self generality)
	ifTrue: [^fraction ~= arg asFraction ]
	ifFalse: [^self retryInequalityCoercing: arg]!

hash
    "Answer an hash value for the receiver."

    ^fraction hash! !

!ScaledDecimal methodsFor: 'printing'!

displayOn: aStream
    "Print a representation of the receiver on aStream, intended to
     be directed to a user.  In this particular case, the `scale'
     part of the #printString is not emitted."
    | aFraction fracDigits |
    self < 0 ifTrue: [aStream nextPut: $-].
    aFraction := fraction abs.
    aStream nextPutAll: aFraction truncated printString.
    scale = 0 ifTrue: [^self].
    aStream nextPut: $. .
    fracDigits := aFraction fractionPart.
    scale timesRepeat: [
	fracDigits := fracDigits * 10.
	aStream nextPut: (Character digitValue: fracDigits truncated).
	fracDigits := fracDigits fractionPart
    ]!

printOn: aStream 
    "Print a representation of the receiver on aStream."
    self displayOn: aStream.
    aStream nextPut: $s.
    scale printOn: aStream! !

!ScaledDecimal methodsFor: 'storing'!

storeOn: aStream 
    "Print Smalltalk code that compiles to the receiver on aStream."
    self printOn: aStream! !

!ScaledDecimal methodsFor: 'private'!

scale
    "Private - Answer a integer which represents the total number of digits 
    used to represent the fraction part of the receiver, including trailing 
    zeroes."
    ^scale!

setFraction: theFraction scale: theScale
    "Private - Set the fraction to theFraction and the total number of digits 
    used to represent the fraction part of the receiver, including trailing 
    zeroes, to the Integer theScale."
    fraction := theFraction.
    scale := theScale! !

ScaledDecimal initialize!
